# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#

# A parser simulator that mimics what ANTLR's generated
#  parser code does. A ParserATNSimulator is used to make
#  predictions via adaptivePredict but this class moves a pointer through the
#  ATN to simulate parsing. ParserATNSimulator just
#  makes us efficient rather than having to backtrack, for example.
#
#  This properly creates parse trees even for left recursive rules.
#
#  We rely on the left recursive rule invocation and special predicate
#  transitions to make left recursive rules work.
#
#  See TestParserInterpreter for examples.
#
from .dfa.DFA import DFA
from .BufferedTokenStream import TokenStream
from .Lexer import Lexer
from .Parser import Parser
from .ParserRuleContext import InterpreterRuleContext, ParserRuleContext
from .Token import Token
from .atn.ATN import ATN
from .atn.ATNState import StarLoopEntryState, ATNState, LoopEndState
from .atn.ParserATNSimulator import ParserATNSimulator
from .PredictionContext import PredictionContextCache
from .atn.Transition import Transition
from .error.Errors import RecognitionException, UnsupportedOperationException, FailedPredicateException


class ParserInterpreter(Parser):
    __slots__ = (
        'grammarFileName', 'atn', 'tokenNames', 'ruleNames', 'decisionToDFA',
        'sharedContextCache', '_parentContextStack',
        'pushRecursionContextStates'
    )

    def __init__(self, grammarFileName:str, tokenNames:list, ruleNames:list, atn:ATN, input:TokenStream):
        super().__init__(input)
        self.grammarFileName = grammarFileName
        self.atn = atn
        self.tokenNames = tokenNames
        self.ruleNames = ruleNames
        self.decisionToDFA = [ DFA(state) for state in atn.decisionToState ]
        self.sharedContextCache = PredictionContextCache()
        self._parentContextStack = list()
        # identify the ATN states where pushNewRecursionContext must be called
        self.pushRecursionContextStates = set()
        for state in atn.states:
            if not isinstance(state, StarLoopEntryState):
                continue
            if state.isPrecedenceDecision:
                self.pushRecursionContextStates.add(state.stateNumber)
        # get atn simulator that knows how to do predictions
        self._interp = ParserATNSimulator(self, atn, self.decisionToDFA, self.sharedContextCache)

    # Begin parsing at startRuleIndex#
    def parse(self, startRuleIndex:int):
        startRuleStartState = self.atn.ruleToStartState[startRuleIndex]
        rootContext = InterpreterRuleContext(None, ATNState.INVALID_STATE_NUMBER, startRuleIndex)
        if startRuleStartState.isPrecedenceRule:
            self.enterRecursionRule(rootContext, startRuleStartState.stateNumber, startRuleIndex, 0)
        else:
            self.enterRule(rootContext, startRuleStartState.stateNumber, startRuleIndex)
        while True:
            p = self.getATNState()
            if p.stateType==ATNState.RULE_STOP :
                # pop; return from rule
                if len(self._ctx)==0:
                    if startRuleStartState.isPrecedenceRule:
                        result = self._ctx
                        parentContext = self._parentContextStack.pop()
                        self.unrollRecursionContexts(parentContext.a)
                        return result
                    else:
                        self.exitRule()
                        return rootContext
                self.visitRuleStopState(p)

            else:
                try:
                    self.visitState(p)
                except RecognitionException as e:
                    self.state = self.atn.ruleToStopState[p.ruleIndex].stateNumber
                    self._ctx.exception = e
                    self._errHandler.reportError(self, e)
                    self._errHandler.recover(self, e)

    def enterRecursionRule(self, localctx:ParserRuleContext, state:int, ruleIndex:int, precedence:int):
        self._parentContextStack.append((self._ctx, localctx.invokingState))
        super().enterRecursionRule(localctx, state, ruleIndex, precedence)

    def getATNState(self):
        return self.atn.states[self.state]

    def visitState(self, p:ATNState):
        edge = 0
        if len(p.transitions) > 1:
            self._errHandler.sync(self)
            edge = self._interp.adaptivePredict(self._input, p.decision, self._ctx)
        else:
            edge = 1

        transition = p.transitions[edge - 1]
        tt = transition.serializationType
        if tt==Transition.EPSILON:

            if self.pushRecursionContextStates[p.stateNumber] and not isinstance(transition.target, LoopEndState):
                t = self._parentContextStack[-1]
                ctx = InterpreterRuleContext(t[0], t[1], self._ctx.ruleIndex)
                self.pushNewRecursionContext(ctx, self.atn.ruleToStartState[p.ruleIndex].stateNumber, self._ctx.ruleIndex)

        elif tt==Transition.ATOM:

            self.match(transition.label)

        elif tt in [ Transition.RANGE, Transition.SET, Transition.NOT_SET]:

            if not transition.matches(self._input.LA(1), Token.MIN_USER_TOKEN_TYPE, Lexer.MAX_CHAR_VALUE):
                self._errHandler.recoverInline(self)
            self.matchWildcard()

        elif tt==Transition.WILDCARD:

            self.matchWildcard()

        elif tt==Transition.RULE:

            ruleStartState = transition.target
            ruleIndex = ruleStartState.ruleIndex
            ctx = InterpreterRuleContext(self._ctx, p.stateNumber, ruleIndex)
            if ruleStartState.isPrecedenceRule:
                self.enterRecursionRule(ctx, ruleStartState.stateNumber, ruleIndex, transition.precedence)
            else:
                self.enterRule(ctx, transition.target.stateNumber, ruleIndex)

        elif tt==Transition.PREDICATE:

            if not self.sempred(self._ctx, transition.ruleIndex, transition.predIndex):
                raise FailedPredicateException(self)

        elif tt==Transition.ACTION:

            self.action(self._ctx, transition.ruleIndex, transition.actionIndex)

        elif tt==Transition.PRECEDENCE:

            if not self.precpred(self._ctx, transition.precedence):
                msg = "precpred(_ctx, " + str(transition.precedence) + ")"
                raise FailedPredicateException(self, msg)

        else:
            raise UnsupportedOperationException("Unrecognized ATN transition type.")

        self.state = transition.target.stateNumber

    def visitRuleStopState(self, p:ATNState):
        ruleStartState = self.atn.ruleToStartState[p.ruleIndex]
        if ruleStartState.isPrecedenceRule:
            parentContext = self._parentContextStack.pop()
            self.unrollRecursionContexts(parentContext.a)
            self.state = parentContext[1]
        else:
            self.exitRule()

        ruleTransition = self.atn.states[self.state].transitions[0]
        self.state = ruleTransition.followState.stateNumber



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
# Based on bluelet ( https://github.com/sampsyo/bluelet ) by Adrian Sampson, original license:
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import dataclasses as dc
import socket
import typing as ta

from omlish.lite.abstract import Abstract

from .core import BlueletCoro
from .core import ReturnBlueletEvent
from .core import ValueBlueletEvent
from .core import _CoreBlueletApi
from .events import BlueletEvent
from .events import BlueletWaitables
from .events import WaitableBlueletEvent


##


class SocketClosedBlueletError(Exception):
    pass


class BlueletListener:
    """A socket wrapper object for listening sockets."""

    def __init__(self, host: str, port: int) -> None:
        """Create a listening socket on the given hostname and port."""

        super().__init__()

        self._closed = False
        self.host = host
        self.port = port

        self.sock = sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((host, port))
        sock.listen(5)

    def accept(self) -> 'AcceptBlueletEvent':
        """
        An event that waits for a connection on the listening socket. When a connection is made, the event returns a
        Connection object.
        """

        if self._closed:
            raise SocketClosedBlueletError
        return AcceptBlueletEvent(self)

    def close(self) -> None:
        """Immediately close the listening socket. (Not an event.)"""

        self._closed = True
        self.sock.close()


class BlueletConnection:
    """A socket wrapper object for connected sockets."""

    def __init__(self, sock: socket.socket, addr: ta.Tuple[str, int]) -> None:
        super().__init__()

        self.sock = sock
        self.addr = addr
        self._buf = bytearray()
        self._closed: bool = False

    def close(self) -> None:
        """Close the connection."""

        self._closed = True
        self.sock.close()

    def recv(self, size: int) -> BlueletEvent:
        """Read at most size bytes of data from the socket."""

        if self._closed:
            raise SocketClosedBlueletError

        if self._buf:
            # We already have data read previously.
            out = self._buf[:size]
            self._buf = self._buf[size:]
            return ValueBlueletEvent(bytes(out))
        else:
            return ReceiveBlueletEvent(self, size)

    def send(self, data: bytes) -> BlueletEvent:
        """Sends data on the socket, returning the number of bytes successfully sent."""

        if self._closed:
            raise SocketClosedBlueletError
        return SendBlueletEvent(self, data)

    def sendall(self, data: bytes) -> BlueletEvent:
        """Send all of data on the socket."""

        if self._closed:
            raise SocketClosedBlueletError
        return SendBlueletEvent(self, data, True)

    def readline(self, terminator: bytes = b'\n', bufsize: int = 1024) -> BlueletCoro:
        """Reads a line (delimited by terminator) from the socket."""

        if self._closed:
            raise SocketClosedBlueletError

        while True:
            if terminator in self._buf:
                line, self._buf = self._buf.split(terminator, 1)
                line += terminator
                yield ReturnBlueletEvent(bytes(line))
                break

            if (data := (yield ReceiveBlueletEvent(self, bufsize))):
                self._buf += data
            else:
                line = self._buf
                self._buf = bytearray()
                yield ReturnBlueletEvent(bytes(line))
                break


##


class SocketBlueletEvent(BlueletEvent, Abstract):
    pass


@dc.dataclass(frozen=True, eq=False)
class AcceptBlueletEvent(WaitableBlueletEvent, SocketBlueletEvent):
    """An event for Listener objects (listening sockets) that suspends execution until the socket gets a connection."""

    listener: BlueletListener

    def waitables(self) -> BlueletWaitables:
        return BlueletWaitables(r=[self.listener.sock])

    def fire(self) -> BlueletConnection:
        sock, addr = self.listener.sock.accept()
        return BlueletConnection(sock, addr)


@dc.dataclass(frozen=True, eq=False)
class ReceiveBlueletEvent(WaitableBlueletEvent, SocketBlueletEvent):
    """An event for Connection objects (connected sockets) for asynchronously reading data."""

    conn: BlueletConnection
    bufsize: int

    def waitables(self) -> BlueletWaitables:
        return BlueletWaitables(r=[self.conn.sock])

    def fire(self) -> bytes:
        return self.conn.sock.recv(self.bufsize)


@dc.dataclass(frozen=True, eq=False)
class SendBlueletEvent(WaitableBlueletEvent, SocketBlueletEvent):
    """An event for Connection objects (connected sockets) for asynchronously writing data."""

    conn: BlueletConnection
    data: bytes
    sendall: bool = False

    def waitables(self) -> BlueletWaitables:
        return BlueletWaitables(w=[self.conn.sock])

    def fire(self) -> ta.Optional[int]:
        if self.sendall:
            self.conn.sock.sendall(self.data)
            return None
        else:
            return self.conn.sock.send(self.data)


##


class _SocketsBlueletApi(_CoreBlueletApi):
    def connect(self, host: str, port: int) -> BlueletEvent:
        """Event: connect to a network address and return a Connection object for communicating on the socket."""

        addr = (host, port)
        sock = socket.create_connection(addr)
        return ValueBlueletEvent(BlueletConnection(sock, addr))

    def server(self, host: str, port: int, func) -> BlueletCoro:
        """
        A coroutine that runs a network server. Host and port specify the listening address. func should be a coroutine
        that takes a single parameter, a Connection object. The coroutine is invoked for every incoming connection on
        the listening socket.
        """

        def handler(conn):
            try:
                yield func(conn)
            finally:
                conn.close()

        listener = BlueletListener(host, port)
        try:
            while True:
                conn = yield listener.accept()
                yield self.spawn(handler(conn))
        except KeyboardInterrupt:
            pass
        finally:
            listener.close()



import dataclasses as dc
import typing as ta

from omlish import collections as col


##


@dc.dataclass(frozen=True)
class NumericReply:
    name: str
    num: int
    formats: ta.Sequence[str]

    @classmethod
    def new(
            cls,
            name: str,
            num: int,
            *formats: str,
    ) -> 'NumericReply':
        return cls(
            name,
            num,
            formats,
        )


##


class NumericReplies:
    def __init__(self, lst: ta.Iterable[NumericReply]) -> None:
        super().__init__()

        self._lst = list(lst)
        self._by_name = col.make_map_by(lambda nr: nr.name, self._lst, strict=True)
        self._by_num = col.make_map_by(lambda nr: nr.num, self._lst, strict=True)

    def __len__(self) -> int:
        return len(self._lst)

    def __iter__(self) -> ta.Iterator[NumericReply]:
        return iter(self._lst)

    def __getitem__(self, key: str | int) -> NumericReply:
        if isinstance(key, str):
            return self._by_name[key]
        elif isinstance(key, int):
            return self._by_num[key]
        else:
            raise TypeError(key)

    def get(self, key: str | int) -> NumericReply | None:
        try:
            return self[key]
        except KeyError:
            return None



import typing as ta

from omlish.graphs import dot

from . import runtime as antlr4
from .utils import yield_contexts


##


def dot_ctx(
        root: antlr4.ParserRuleContext,
        *,
        left_to_right: bool = False,
) -> dot.Graph:
    stmts: list[dot.Stmt] = []

    if left_to_right:
        stmts.append(dot.RawStmt('rankdir=LR;'))

    for c in yield_contexts(root):
        if isinstance(c, antlr4.TerminalNode):
            continue

        lbl = [
            [type(c).__name__],
            [str(id(c))],
            [f'{c.start} {c.stop}'],
        ]

        stmts.append(dot.Node(f'_{id(c)}', {'label': lbl, 'shape': 'box'}))

        for n in (c.children or []):
            if not isinstance(n, antlr4.TerminalNode):
                stmts.append(dot.Edge(f'_{id(c)}', f'_{id(n)}'))

    return dot.Graph(stmts)


def open_dot_ctx(root: antlr4.ParserRuleContext, **kwargs: ta.Any) -> None:
    dot.open_dot(dot.render(dot_ctx(root)), **kwargs)



import typing as ta

from omlish import check
from omlish import dataclasses as dc
from omlish.funcs import pairs as fps

from ..numerics import numerics as nr
from .formats import MessageFormat
from .formats import MessageParamsUnpacker


##


class Message(dc.Case):
    FORMAT: ta.ClassVar[MessageFormat]
    REPLIES: ta.ClassVar[ta.Collection[nr.NumericReply]] = ()


##


def list_pair_params_unpacker(
        kwarg: str,
        key_param: str,
        value_param: str,
) -> MessageParamsUnpacker:
    def forward(params: ta.Mapping[str, str]) -> ta.Mapping[str, ta.Any]:
        out: dict = dict(params)
        ks = out.pop(key_param)
        vs = out.pop(value_param, None)
        if vs is not None:
            out[kwarg] = list(zip(ks, vs, strict=True))
        else:
            out[kwarg] = ks
        return out

    def backward(kwargs: ta.Mapping[str, ta.Any]) -> ta.Mapping[str, str]:
        out: dict = dict(kwargs)
        ts = out.pop(kwarg)
        is_ts = check.single({isinstance(e, tuple) for e in ts})
        if is_ts:
            ks, vs = zip(*ts)
            out[key_param] = ks
            out[value_param] = vs
        else:
            out[key_param] = ts
        return out

    return fps.of(forward, backward)



# ruff: noqa: N802 N815
import io
import typing as ta

from omlish import check

from . import runtime as antlr4


##


class DelimitingLexer(antlr4.Lexer):
    def __init__(
            self,
            *args: ta.Any,
            delimiter_token: ta.Any,
            delimiters: ta.Iterable[str],
            no_skip: bool = False,
            **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self._delimiter_token = delimiter_token
        self._delimiters = set(check.not_isinstance(delimiters, str))
        self._no_skip = no_skip

    _hitEOF: bool

    def nextToken(self) -> antlr4.Token:
        if self._input is None:
            raise antlr4.IllegalStateException('nextToken requires a non-null input stream.')

        token_start_marker = self._input.mark()
        try:
            while True:
                if self._hitEOF:
                    self.emitEOF()
                    return self._token

                self._token: antlr4.Token | None = None
                self._channel = antlr4.Token.DEFAULT_CHANNEL
                self._tokenStartCharIndex = self._input.index
                self._tokenStartColumn = self._interp.column
                self._tokenStartLine = self._interp.line
                self._text = None

                continue_outer = False
                while True:
                    self._type = antlr4.Token.INVALID_TYPE
                    ttype = self.SKIP

                    for delimiter in self._delimiters:
                        if self._match_delimiter(delimiter):
                            ttype = self._delimiter_token
                            break
                    else:
                        try:
                            ttype = self._interp.match(self._input, self._mode)
                        except antlr4.LexerNoViableAltException as e:
                            self.notifyListeners(e)  # report error
                            self.recover(e)

                    if self._input.LA(1) == antlr4.Token.EOF:
                        self._hitEOF = True

                    if self._type == antlr4.Token.INVALID_TYPE:
                        self._type = ttype

                    if not self._no_skip and self._type == self.SKIP:
                        continue_outer = True
                        break

                    if self._type != self.MORE:
                        break

                if continue_outer:
                    continue

                if self._token is None:
                    self.emit()

                return self._token

        finally:
            self._input.release(token_start_marker)

    def _match_delimiter(self, delimiter: str) -> bool:
        for i, c in enumerate(delimiter):
            if chr(self._input.LA(i + 1)) != c:
                return False
        self._input.seek(self._input.index + len(delimiter))
        return True

    def split(self) -> tuple[list[tuple[str, str]], str]:
        lst = []
        sb = io.StringIO()
        while True:
            token = self.nextToken()
            if token.type == antlr4.Token.EOF:
                break
            if token.type == self._delimiter_token:
                statement = sb.getvalue().strip()
                if statement:
                    lst.append((statement, token.text))
                sb = io.StringIO()
            else:
                sb.write(token.text)
        partial = sb.getvalue()
        return lst, partial



from omlish.__about__ import ProjectBase
from omlish.__about__ import SetuptoolsBase
from omlish.__about__ import __version__


class Project(ProjectBase):
    name = 'omextra'
    description = 'omextra'

    dependencies = [
        # FIXME: text.antlr.cli deps omdev.cache.data, yet this lib is 'under' omdev.
        # f'omdev == {__version__}',

        f'omlish == {__version__}',
    ]

    optional_dependencies: dict = {
    }

    entry_points = {
        'omlish.manifests': {name: name},
    }


class Setuptools(SetuptoolsBase):
    find_packages = {
        'include': [Project.name, f'{Project.name}.*'],
        'exclude': [*SetuptoolsBase.find_packages['exclude']],
    }



"""
A self-awarely unpythonic set of helpers for defining common boilerplate methods (repr, hash_eq, delegates, etc) in
class definitions. Should be used sparingly for methods not directly used by humans (like repr) - @property's should
remain @property's for type annotation, tool assistance, debugging, and otherwise, but these are still nice to have in
certain circumstances (the real-world alternative usually being simply not adding them).
"""
# ruff: noqa: ANN201
import abc
import functools
import operator
import threading

from omlish import lang


##


_repr = repr

BASICS = {}


def _basic(fn):
    if fn.__name__ in BASICS:
        raise NameError(fn.__name__)

    BASICS[fn.__name__] = fn
    return fn


@lang.cls_dct_fn()
def basic(cls_dct, *attrs, basics=None):
    if basics is None:
        basics = BASICS.keys()

    for k in basics:
        fn = BASICS[k]
        fn(*attrs, cls_dct=cls_dct)


_REPR_SEEN = threading.local()


def _repr_guard(fn):
    @functools.wraps(fn)
    def inner(obj, *args, **kwargs):
        try:
            ids = _REPR_SEEN.ids

        except AttributeError:
            ids = _REPR_SEEN.ids = set()
            try:
                ids.add(id(obj))
                return fn(obj, *args, **kwargs)
            finally:
                del _REPR_SEEN.ids

        else:
            if id(obj) in ids:
                return f'<seen:{type(obj).__name__}@{id(obj):x}>'
            ids.add(id(obj))
            return fn(obj, *args, **kwargs)

    return inner


@_repr_guard
def build_attr_repr(obj, *, mro=False):
    if mro:
        attrs = [
            attr
            for ty in sorted(  # noqa
                reversed(type(obj).__mro__),
                key=lambda _ty: _ty.__dict__.get('__repr_priority__', 0),
            )
            for attr in ty.__dict__.get('__repr_attrs__', [])
        ]

    else:
        attrs = obj.__repr_attrs__

    s = ', '.join(f'{a}={"<self>" if v is obj else _repr(v)}' for a in attrs for v in [getattr(obj, a)])
    return f'{type(obj).__name__}@{id(obj):x}({s})'


@_repr_guard
def build_repr(obj, *attrs):
    return (
        f'{type(obj).__name__}'
        f'@{id(obj):x}'
        f'({", ".join(f"{attr}={getattr(obj, attr)!r}" for attr in attrs)})'
    )


@_basic
@lang.cls_dct_fn()
def repr(cls_dct, *attrs, mro=False, priority=None):  # noqa
    def __repr__(self):  # noqa
        return build_attr_repr(self, mro=mro)

    cls_dct['__repr_attrs__'] = attrs
    if priority is not None:
        cls_dct['__repr_priority__'] = priority

    cls_dct['__repr__'] = __repr__


@lang.cls_dct_fn()
def bare_repr(cls_dct, *attrs):
    def __repr__(self):  # noqa
        return lang.attr_repr(self, *attrs)

    cls_dct['__repr__'] = __repr__


@lang.cls_dct_fn()
def name_repr(cls_dct):
    def __repr__(self):  # noqa
        return self.__name__

    cls_dct['__repr__'] = __repr__


@lang.cls_dct_fn()
def ne(cls_dct):
    def __ne__(self, other):  # noqa
        return not (self == other)

    cls_dct['__ne__'] = __ne__


@lang.cls_dct_fn()
def no_order(cls_dct, *, raise_=None):
    def fn(self, other):
        if raise_ is None:
            return NotImplemented
        else:
            raise raise_

    for att in [
        '__lt__',
        '__le__',
        '__gt__',
        '__ge__',
    ]:
        cls_dct[att] = fn


@_basic
@lang.cls_dct_fn()
def hash_eq(cls_dct, *attrs):
    def __hash__(self):  # noqa
        return hash(tuple(getattr(self, attr) for attr in attrs))

    cls_dct['__hash__'] = __hash__

    def __eq__(self, other):  # noqa
        if type(other) is not type(self):
            return False

        for attr in attrs:  # noqa
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    cls_dct['__eq__'] = __eq__

    ne(cls_dct=cls_dct)


@lang.cls_dct_fn()
def getter(cls_dct, *attrs):
    for attr in attrs:
        cls_dct[attr] = property(operator.attrgetter('_' + attr))


@lang.cls_dct_fn()
def not_implemented(cls_dct, *names, **kwargs):
    wrapper = kwargs.pop('wrapper', lambda _: _)
    if kwargs:
        raise TypeError(kwargs)

    ret = []
    for name in names:
        @wrapper
        def not_implemented(self, *args, **kwargs):
            raise NotImplementedError

        not_implemented.__name__ = name
        cls_dct[name] = not_implemented
        ret.append(not_implemented)

    return tuple(ret)


@lang.cls_dct_fn()
def abstract_method(cls_dct, *names):
    return not_implemented(cls_dct, *names, wrapper=abc.abstractmethod)


@lang.cls_dct_fn()
def abstract_property(cls_dct, *names):
    return not_implemented(cls_dct, *names, wrapper=lambda o: property(abc.abstractmethod(o)))


@lang.cls_dct_fn()
def abstract_hash_eq(cls_dct):
    return not_implemented(
        cls_dct,
        '__hash__',
        '__eq__',
        '__ne__',
        wrapper=abc.abstractmethod,
    )



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#

class Chunk(object):
    pass

class TagChunk(Chunk):
    __slots__ = ('tag', 'label')

    def __init__(self, tag:str, label:str=None):
        self.tag = tag
        self.label = label

    def __str__(self):
        if self.label is None:
            return self.tag
        else:
            return self.label + ":" + self.tag

class TextChunk(Chunk):
    __slots__ = 'text'

    def __init__(self, text:str):
        self.text = text

    def __str__(self):
        return "'" + self.text + "'"



# ruff: noqa: N802 N803
import typing as ta

from omlish.formats.json5.errors import Json5Error
from omlish.formats.json5.literals import LITERAL_VALUES
from omlish.formats.json5.literals import parse_number_literal
from omlish.formats.json5.literals import parse_string_literal

from ...text import antlr
from ._antlr.Json5Lexer import Json5Lexer  # type: ignore
from ._antlr.Json5Parser import Json5Parser  # type: ignore
from ._antlr.Json5Visitor import Json5Visitor  # type: ignore


##


class Json5ParseVisitor(antlr.parsing.StandardParseTreeVisitor, Json5Visitor):
    def visitArr(self, ctx: Json5Parser.ArrContext):
        return [self.visit(e) for e in ctx.value()]

    def visitKey(self, ctx: Json5Parser.KeyContext):
        if (s := ctx.STRING()) is not None:
            return parse_string_literal(s.getText())

        elif (i := ctx.IDENTIFIER()) is not None:
            return parse_string_literal(''.join(['"', i.getText(), '"']))

        elif (l := ctx.LITERAL()) is not None:
            return LITERAL_VALUES[l.getText()]

        elif (n := ctx.NUMERIC_LITERAL()) is not None:
            return n.getText()

        else:
            raise RuntimeError(ctx)

    def visitNumber(self, ctx: Json5Parser.NumberContext):
        return parse_number_literal(ctx.getText())

    def visitObj(self, ctx: Json5Parser.ObjContext):
        dct: dict[ta.Any, ta.Any] = {}
        for pair in ctx.pair():
            key, value = self.visit(pair)
            dct[key] = value
        return dct

    def visitPair(self, ctx: Json5Parser.PairContext):
        key = self.visit(ctx.key())
        value = self.visit(ctx.value())
        return (key, value)

    def visitValue(self, ctx: Json5Parser.ValueContext):
        if (s := ctx.STRING()) is not None:
            return parse_string_literal(s.getText())

        elif (n := ctx.LITERAL()) is not None:
            return LITERAL_VALUES[n.getText()]

        else:
            return super().visitChildren(ctx)


def _make_parser(buf: str) -> Json5Parser:
    return antlr.parsing.make_parser(
        buf,
        Json5Lexer,
        Json5Parser,
        silent_errors=True,
    )


def parse(buf: str) -> ta.Any:
    try:
        root = _make_parser(buf).json5()

    except antlr.errors.ParseError as e:
        raise Json5Error from e

    if antlr.parsing.is_eof_context(root):
        raise Json5Error('Empty input')

    visitor = Json5ParseVisitor()
    return visitor.visit(root)


def parse_many(buf: str) -> ta.Iterator[ta.Any]:
    try:
        parser = _make_parser(buf)

        while True:
            if parser.getInputStream().LT(1).type == antlr.runtime.Token.EOF:
                break

            value = parser.value()

            visitor = Json5ParseVisitor()
            yield visitor.visit(value)

    except antlr.errors.ParseError as e:
        raise Json5Error from e



# type: ignore
# ruff: noqa
# flake8: noqa
# @omlish-generated
# Generated from Json.g4 by ANTLR 4.13.2
# encoding: utf-8
from ....text.antlr._runtime._all import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,12,57,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,1,0,1,0,1,0,1,
        1,1,1,1,1,1,1,5,1,18,8,1,10,1,12,1,21,9,1,1,1,1,1,1,1,1,1,3,1,27,
        8,1,1,2,1,2,1,2,1,2,1,3,1,3,1,3,1,3,5,3,37,8,3,10,3,12,3,40,9,3,
        1,3,1,3,1,3,1,3,3,3,46,8,3,1,4,1,4,1,4,1,4,1,4,1,4,1,4,3,4,55,8,
        4,1,4,0,0,5,0,2,4,6,8,0,0,61,0,10,1,0,0,0,2,26,1,0,0,0,4,28,1,0,
        0,0,6,45,1,0,0,0,8,54,1,0,0,0,10,11,3,8,4,0,11,12,5,0,0,1,12,1,1,
        0,0,0,13,14,5,1,0,0,14,19,3,4,2,0,15,16,5,2,0,0,16,18,3,4,2,0,17,
        15,1,0,0,0,18,21,1,0,0,0,19,17,1,0,0,0,19,20,1,0,0,0,20,22,1,0,0,
        0,21,19,1,0,0,0,22,23,5,3,0,0,23,27,1,0,0,0,24,25,5,1,0,0,25,27,
        5,3,0,0,26,13,1,0,0,0,26,24,1,0,0,0,27,3,1,0,0,0,28,29,5,10,0,0,
        29,30,5,4,0,0,30,31,3,8,4,0,31,5,1,0,0,0,32,33,5,5,0,0,33,38,3,8,
        4,0,34,35,5,2,0,0,35,37,3,8,4,0,36,34,1,0,0,0,37,40,1,0,0,0,38,36,
        1,0,0,0,38,39,1,0,0,0,39,41,1,0,0,0,40,38,1,0,0,0,41,42,5,6,0,0,
        42,46,1,0,0,0,43,44,5,5,0,0,44,46,5,6,0,0,45,32,1,0,0,0,45,43,1,
        0,0,0,46,7,1,0,0,0,47,55,5,10,0,0,48,55,5,11,0,0,49,55,3,2,1,0,50,
        55,3,6,3,0,51,55,5,7,0,0,52,55,5,8,0,0,53,55,5,9,0,0,54,47,1,0,0,
        0,54,48,1,0,0,0,54,49,1,0,0,0,54,50,1,0,0,0,54,51,1,0,0,0,54,52,
        1,0,0,0,54,53,1,0,0,0,55,9,1,0,0,0,5,19,26,38,45,54
    ]

class JsonParser ( Parser ):

    grammarFileName = "Json.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'{'", "','", "'}'", "':'", "'['", "']'", 
                     "'true'", "'false'", "'null'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "STRING", "NUMBER", "WS" ]

    RULE_json = 0
    RULE_obj = 1
    RULE_pair = 2
    RULE_arr = 3
    RULE_value = 4

    ruleNames =  [ "json", "obj", "pair", "arr", "value" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    STRING=10
    NUMBER=11
    WS=12

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class JsonContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def value(self):
            return self.getTypedRuleContext(JsonParser.ValueContext,0)


        def EOF(self):
            return self.getToken(JsonParser.EOF, 0)

        def getRuleIndex(self):
            return JsonParser.RULE_json

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterJson" ):
                listener.enterJson(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitJson" ):
                listener.exitJson(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitJson" ):
                return visitor.visitJson(self)
            else:
                return visitor.visitChildren(self)




    def json(self):

        localctx = JsonParser.JsonContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_json)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 10
            self.value()
            self.state = 11
            self.match(JsonParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ObjContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def pair(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JsonParser.PairContext)
            else:
                return self.getTypedRuleContext(JsonParser.PairContext,i)


        def getRuleIndex(self):
            return JsonParser.RULE_obj

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObj" ):
                listener.enterObj(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObj" ):
                listener.exitObj(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObj" ):
                return visitor.visitObj(self)
            else:
                return visitor.visitChildren(self)




    def obj(self):

        localctx = JsonParser.ObjContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_obj)
        self._la = 0 # Token type
        try:
            self.state = 26
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 13
                self.match(JsonParser.T__0)
                self.state = 14
                self.pair()
                self.state = 19
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==2:
                    self.state = 15
                    self.match(JsonParser.T__1)
                    self.state = 16
                    self.pair()
                    self.state = 21
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 22
                self.match(JsonParser.T__2)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 24
                self.match(JsonParser.T__0)
                self.state = 25
                self.match(JsonParser.T__2)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PairContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(JsonParser.STRING, 0)

        def value(self):
            return self.getTypedRuleContext(JsonParser.ValueContext,0)


        def getRuleIndex(self):
            return JsonParser.RULE_pair

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPair" ):
                listener.enterPair(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPair" ):
                listener.exitPair(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPair" ):
                return visitor.visitPair(self)
            else:
                return visitor.visitChildren(self)




    def pair(self):

        localctx = JsonParser.PairContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_pair)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 28
            self.match(JsonParser.STRING)
            self.state = 29
            self.match(JsonParser.T__3)
            self.state = 30
            self.value()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ArrContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def value(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JsonParser.ValueContext)
            else:
                return self.getTypedRuleContext(JsonParser.ValueContext,i)


        def getRuleIndex(self):
            return JsonParser.RULE_arr

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterArr" ):
                listener.enterArr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitArr" ):
                listener.exitArr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitArr" ):
                return visitor.visitArr(self)
            else:
                return visitor.visitChildren(self)




    def arr(self):

        localctx = JsonParser.ArrContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_arr)
        self._la = 0 # Token type
        try:
            self.state = 45
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,3,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 32
                self.match(JsonParser.T__4)
                self.state = 33
                self.value()
                self.state = 38
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==2:
                    self.state = 34
                    self.match(JsonParser.T__1)
                    self.state = 35
                    self.value()
                    self.state = 40
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 41
                self.match(JsonParser.T__5)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 43
                self.match(JsonParser.T__4)
                self.state = 44
                self.match(JsonParser.T__5)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ValueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(JsonParser.STRING, 0)

        def NUMBER(self):
            return self.getToken(JsonParser.NUMBER, 0)

        def obj(self):
            return self.getTypedRuleContext(JsonParser.ObjContext,0)


        def arr(self):
            return self.getTypedRuleContext(JsonParser.ArrContext,0)


        def getRuleIndex(self):
            return JsonParser.RULE_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterValue" ):
                listener.enterValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitValue" ):
                listener.exitValue(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitValue" ):
                return visitor.visitValue(self)
            else:
                return visitor.visitChildren(self)




    def value(self):

        localctx = JsonParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_value)
        try:
            self.state = 54
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [10]:
                self.enterOuterAlt(localctx, 1)
                self.state = 47
                self.match(JsonParser.STRING)
                pass
            elif token in [11]:
                self.enterOuterAlt(localctx, 2)
                self.state = 48
                self.match(JsonParser.NUMBER)
                pass
            elif token in [1]:
                self.enterOuterAlt(localctx, 3)
                self.state = 49
                self.obj()
                pass
            elif token in [5]:
                self.enterOuterAlt(localctx, 4)
                self.state = 50
                self.arr()
                pass
            elif token in [7]:
                self.enterOuterAlt(localctx, 5)
                self.state = 51
                self.match(JsonParser.T__6)
                pass
            elif token in [8]:
                self.enterOuterAlt(localctx, 6)
                self.state = 52
                self.match(JsonParser.T__7)
                pass
            elif token in [9]:
                self.enterOuterAlt(localctx, 7)
                self.state = 53
                self.match(JsonParser.T__8)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
# Based on bluelet ( https://github.com/sampsyo/bluelet ) by Adrian Sampson, original license:
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import dataclasses as dc
import typing as ta

from omlish.lite.abstract import Abstract

from .core import DelegationBlueletEvent
from .core import ReturnBlueletEvent
from .events import BlueletEvent
from .events import BlueletWaitables
from .events import WaitableBlueletEvent


##


class FileBlueletEvent(BlueletEvent, Abstract):
    pass


@dc.dataclass(frozen=True, eq=False)
class ReadBlueletEvent(WaitableBlueletEvent, FileBlueletEvent):
    """Reads from a file-like object."""

    fd: ta.IO
    bufsize: int

    def waitables(self) -> BlueletWaitables:
        return BlueletWaitables(r=[self.fd])

    def fire(self) -> bytes:
        return self.fd.read(self.bufsize)


@dc.dataclass(frozen=True, eq=False)
class WriteBlueletEvent(WaitableBlueletEvent, FileBlueletEvent):
    """Writes to a file-like object."""

    fd: ta.IO
    data: bytes

    def waitables(self) -> BlueletWaitables:
        return BlueletWaitables(w=[self.fd])

    def fire(self) -> None:
        self.fd.write(self.data)


##


class _FilesBlueletApi:
    def read(self, fd: ta.IO, bufsize: ta.Optional[int] = None) -> BlueletEvent:
        """Event: read from a file descriptor asynchronously."""

        if bufsize is None:
            # Read all.
            def reader():
                buf = []
                while True:
                    data = yield self.read(fd, 1024)
                    if not data:
                        break
                    buf.append(data)
                yield ReturnBlueletEvent(''.join(buf))

            return DelegationBlueletEvent(reader())

        else:
            return ReadBlueletEvent(fd, bufsize)

    def write(self, fd: ta.IO, data: bytes) -> BlueletEvent:
        """Event: write to a file descriptor asynchronously."""

        return WriteBlueletEvent(fd, data)



# type: ignore
# ruff: noqa
# flake8: noqa
# @omlish-generated
# Generated from Json5.g4 by ANTLR 4.13.2
from ....text.antlr._runtime._all import *
if "." in __name__:
    from .Json5Parser import Json5Parser
else:
    from Json5Parser import Json5Parser

# This class defines a complete listener for a parse tree produced by Json5Parser.
class Json5Listener(ParseTreeListener):

    # Enter a parse tree produced by Json5Parser#json5.
    def enterJson5(self, ctx:Json5Parser.Json5Context):
        pass

    # Exit a parse tree produced by Json5Parser#json5.
    def exitJson5(self, ctx:Json5Parser.Json5Context):
        pass


    # Enter a parse tree produced by Json5Parser#obj.
    def enterObj(self, ctx:Json5Parser.ObjContext):
        pass

    # Exit a parse tree produced by Json5Parser#obj.
    def exitObj(self, ctx:Json5Parser.ObjContext):
        pass


    # Enter a parse tree produced by Json5Parser#pair.
    def enterPair(self, ctx:Json5Parser.PairContext):
        pass

    # Exit a parse tree produced by Json5Parser#pair.
    def exitPair(self, ctx:Json5Parser.PairContext):
        pass


    # Enter a parse tree produced by Json5Parser#key.
    def enterKey(self, ctx:Json5Parser.KeyContext):
        pass

    # Exit a parse tree produced by Json5Parser#key.
    def exitKey(self, ctx:Json5Parser.KeyContext):
        pass


    # Enter a parse tree produced by Json5Parser#value.
    def enterValue(self, ctx:Json5Parser.ValueContext):
        pass

    # Exit a parse tree produced by Json5Parser#value.
    def exitValue(self, ctx:Json5Parser.ValueContext):
        pass


    # Enter a parse tree produced by Json5Parser#arr.
    def enterArr(self, ctx:Json5Parser.ArrContext):
        pass

    # Exit a parse tree produced by Json5Parser#arr.
    def exitArr(self, ctx:Json5Parser.ArrContext):
        pass


    # Enter a parse tree produced by Json5Parser#number.
    def enterNumber(self, ctx:Json5Parser.NumberContext):
        pass

    # Exit a parse tree produced by Json5Parser#number.
    def exitNumber(self, ctx:Json5Parser.NumberContext):
        pass



del Json5Parser


"""
A system for bridging sync and async code. Supports nesting / reentrancy. Robust, but not the most efficient
implementation - its primary usecase are heavy and init-time ops like lock management and connection establishment, not
so much for packet-level ops.

==

The code is written in a very dumb, simple, explicit style to keep the bookkeeping center-stage and front-of-mind. To
quote kubernetes:

  https://github.com/kubernetes/kubernetes/blob/60c4c2b2521fb454ce69dee737e3eb91a25e0535/pkg/controller/volume/persistentvolume/pv_controller.go#L60-L63

  ==================================================================
  PLEASE DO NOT ATTEMPT TO SIMPLIFY THIS CODE.
  KEEP THE SPACE SHUTTLE FLYING.
  ==================================================================

==

TODO:
 - reuse greenlet if nested somehow?
 - test cancel

See:
 - https://github.com/sqlalchemy/sqlalchemy/blob/21ea01eebe0350ad1185c7288dca61f363ebd2fe/lib/sqlalchemy/util/concurrency.py
   - ( https://gist.github.com/snaury/202bf4f22c41ca34e56297bae5f33fef )
   - Baked into sqlalchemy and hard-bound to asyncio.
 - https://github.com/oremanj/greenback/blob/ca69b023a9b7b58b715f3b1d78fc116e788a2c9f/greenback/_impl.py
   - ( https://gist.github.com/oremanj/f18ef3e55b9487c2e93eee42232583f2 )
   - Similar to this, but does horrible things with ctypes and is hard-bound to greenlet.
 - https://github.com/fastapi/asyncer/blob/2a4b8ef2540ec687af13d3f361c4ed0cf0cb624d/asyncer/_main.py
 - https://github.com/django/asgiref/blob/05ae3eee3fae4005ae4cfb0bb22d281725fabade/asgiref/sync.py
   - Both are much heavier weight and hard-bound to threads.
"""  # noqa
import functools  # noqa
import itertools
import sys  # noqa
import types
import typing as ta
import weakref

from omlish import check
from omlish import lang
from omlish import sync
from omlish.concurrent import threadlets


if ta.TYPE_CHECKING:
    import asyncio

    from omlish.asyncs import anyio as aiu

else:
    asyncio = lang.proxy_import('asyncio')

    aiu = lang.proxy_import('omlish.asyncs.anyio')


T = ta.TypeVar('T')


##


def trivial_s_to_a(fn):
    async def inner(*args, **kwargs):
        return fn(*args, **kwargs)
    return inner


def trivial_a_to_s(fn):
    def inner(*args, **kwargs):
        return lang.sync_await(fn(*args, **kwargs))
    return inner


##
# https://gist.github.com/snaury/202bf4f22c41ca34e56297bae5f33fef


_THREADLETS_IMPL = threadlets.GreenletThreadlets
# from ..concurrent.tests.real import RealThreadlets
# _THREADLETS_IMPL = RealThreadlets

_THREADLETS = sync.LazyFn(lambda: _THREADLETS_IMPL())


def _threadlets() -> threadlets.Threadlets:
    return _THREADLETS.get()


#


class BridgeAwaitRequiredError(Exception):
    pass


class MissingBridgeThreadletError(Exception):
    pass


class UnexpectedBridgeNestingError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # breakpoint()


#


_DEBUG_PRINT: ta.Callable[..., None] | None = None
# _DEBUG_PRINT = functools.partial(print, file=sys.stderr)  # noqa

_TRACK_TRANSITION_OBJS = False


#


_BRIDGE_TRANSITIONS_SEQ = itertools.count()


class _BridgeTransition(ta.NamedTuple):
    seq: int
    a_to_s: bool

    obj_cls: type
    obj_id: int

    obj: ta.Any


def _make_transition(seq: int, a_to_s: bool, obj: ta.Any) -> _BridgeTransition:
    return _BridgeTransition(seq, a_to_s, obj.__class__, id(obj), (obj if _TRACK_TRANSITION_OBJS else None))


_BRIDGED_TASKS: ta.MutableMapping[ta.Any, list[_BridgeTransition]] = weakref.WeakKeyDictionary()

_BRIDGE_THREADLET_ATTR = f'__{__package__.replace(".", "__")}__bridge_threadlet__'


def _push_transition(a_to_s: bool, l: list[_BridgeTransition], t: _BridgeTransition) -> _BridgeTransition:
    l.append(t)
    if _DEBUG_PRINT:
        _DEBUG_PRINT(f'_push_transition: {a_to_s=} {id(l)=} {t=}')
    return t


def _pop_transition(a_to_s: bool, l: list[_BridgeTransition]) -> _BridgeTransition:
    t = l.pop()
    if _DEBUG_PRINT:
        _DEBUG_PRINT(f'_pop_transition: {a_to_s=} {id(l)=} {t=}')
    return t


def _get_transitions() -> list[_BridgeTransition]:
    l: list[_BridgeTransition] = []

    if (t := aiu.get_current_backend_task()) is not None:
        try:
            tl = _BRIDGED_TASKS[t]
        except KeyError:
            pass
        else:
            l.extend(tl)

    g = _threadlets().get_current()
    try:
        gl = getattr(g.underlying, _BRIDGE_THREADLET_ATTR)
    except AttributeError:
        pass
    else:
        l.extend(gl)

    l.sort(key=lambda t: (t.seq, t.a_to_s))
    return l


def is_in_bridge() -> bool:
    if _DEBUG_PRINT:
        _DEBUG_PRINT(_get_transitions())

    if (t := aiu.get_current_backend_task()) is not None:
        try:
            tl = _BRIDGED_TASKS[t]
        except KeyError:
            last_t = None
        else:
            if tl:
                last_t = tl[-1]
            else:
                last_t = None
    else:
        last_t = None

    g = _threadlets().get_current()
    try:
        gl = getattr(g.underlying, _BRIDGE_THREADLET_ATTR)
    except AttributeError:
        last_g = None
    else:
        if gl:
            last_g = gl[-1]
        else:
            last_g = None

    if last_t is None:
        if last_g is None:
            return False
        o = last_g
    else:  # noqa
        if last_g is None or last_g.seq < last_t.seq:
            o = last_t
        else:
            o = last_g

    in_a = (t is not None)

    if _DEBUG_PRINT:
        _DEBUG_PRINT(f'{o.a_to_s=} {in_a=}')

    return in_a != o.a_to_s


def _safe_cancel_awaitable(awaitable: ta.Awaitable[ta.Any]) -> None:
    # https://docs.python.org/3/reference/datamodel.html#coroutine.close
    if asyncio.iscoroutine(awaitable):
        awaitable.close()  # noqa


def s_to_a_await(awaitable: ta.Awaitable[T]) -> T:
    g = _threadlets().get_current()

    if not getattr(g.underlying, _BRIDGE_THREADLET_ATTR, False):
        _safe_cancel_awaitable(awaitable)
        raise MissingBridgeThreadletError

    return check.not_none(g.parent).switch(awaitable)


def s_to_a(fn, *, require_await=False):
    @types.coroutine
    def outer(*args, **kwargs):
        def inner():
            try:
                return fn(*args, **kwargs)
            finally:
                if (gl2 := getattr(g.underlying, _BRIDGE_THREADLET_ATTR)) is not gl:  # noqa
                    raise UnexpectedBridgeNestingError
                if (cur_g := _pop_transition(False, gl)) is not added_g:  # noqa
                    raise UnexpectedBridgeNestingError
                if gl:
                    raise UnexpectedBridgeNestingError

        seq = next(_BRIDGE_TRANSITIONS_SEQ)

        g = _threadlets().spawn(inner)
        setattr(g.underlying, _BRIDGE_THREADLET_ATTR, gl := [])  # type: ignore
        added_g = _push_transition(False, gl, _make_transition(seq, False, g))

        if (t := aiu.get_current_backend_task()) is not None:
            try:
                tl = _BRIDGED_TASKS[t]
            except KeyError:
                tl = _BRIDGED_TASKS[t] = []
            added_t = _push_transition(False, tl, _make_transition(seq, False, g))

        try:
            result: ta.Any = g.switch()
            switch_occurred = False
            while not g.dead:
                switch_occurred = True
                try:
                    value = yield result
                except BaseException as e:  # noqa
                    result = g.throw(e)
                else:
                    result = g.switch(value)

            if require_await and not switch_occurred:
                raise BridgeAwaitRequiredError

            return result

        finally:
            if t is not None:
                if (tl2 := _BRIDGED_TASKS[t]) is not tl:  # noqa
                    raise UnexpectedBridgeNestingError
                if (cur_t := _pop_transition(False, tl)) is not added_t:  # noqa
                    raise UnexpectedBridgeNestingError

    return outer


def a_to_s(fn):
    def inner(*args, **kwargs):
        seq = next(_BRIDGE_TRANSITIONS_SEQ)

        if (t := aiu.get_current_backend_task()) is not None:
            try:
                tl = _BRIDGED_TASKS[t]
            except KeyError:
                tl = _BRIDGED_TASKS[t] = []
            added_t = _push_transition(True, tl, _make_transition(seq, True, t))
        else:
            added_t = None

        g = _threadlets().get_current()
        try:
            gl = getattr(g.underlying, _BRIDGE_THREADLET_ATTR)
        except AttributeError:
            setattr(g.underlying, _BRIDGE_THREADLET_ATTR, gl := [])
        added_g = _push_transition(True, gl, _make_transition(seq, True, g))

        try:
            ret = missing = object()

            async def gate():
                nonlocal ret
                ret = await fn(*args, **kwargs)

            cr = gate()
            sv = None
            he = False
            try:
                while True:
                    if not he:
                        try:
                            sv = cr.send(sv)
                        except StopIteration:
                            break
                    he = False

                    if ret is missing or cr.cr_await is not None or cr.cr_running:
                        try:
                            sv = s_to_a_await(sv)  # type: ignore
                        except BaseException as e:  # noqa
                            sv = cr.throw(e)
                            he = True

            finally:
                cr.close()

        finally:
            if t is not None:
                if (tl2 := _BRIDGED_TASKS[t]) is not tl:  # noqa
                    raise UnexpectedBridgeNestingError
                if (cur_t := _pop_transition(True, tl)) is not added_t:  # noqa
                    raise UnexpectedBridgeNestingError

            if (gl2 := getattr(g.underlying, _BRIDGE_THREADLET_ATTR)) is not gl:  # noqa
                raise UnexpectedBridgeNestingError
            if (cur_g := _pop_transition(True, gl)) is not added_g:  # noqa
                raise UnexpectedBridgeNestingError

        return ret

    return inner



import logging
import re
import subprocess
import sys

from omlish.argparse import all as ap
from omlish.logs import all as logs

from .consts import ANTLR_RUNTIME_PACKAGE
from .gen import GenPy
from .gen import get_jar_path


log = logs.get_module_logger(globals())


##


class Cli(ap.Cli):
    @ap.cmd()
    def jar(self) -> None:
        print(get_jar_path())

    @ap.cmd()
    def latest(self) -> None:
        o, _ = subprocess.Popen(
            [
                sys.executable,
                '-m', 'pip',
                'index', 'versions',
                ANTLR_RUNTIME_PACKAGE,
            ],
            stdout=subprocess.PIPE,
        ).communicate()
        tl = o.decode().splitlines()[0]
        m = re.fullmatch(rf'{ANTLR_RUNTIME_PACKAGE} \((?P<version>[^)]+)\)', tl)
        if m is None:
            raise ValueError(f'Failed to parse version: {tl}')
        v = m.groupdict()['version']
        print(v)

    #

    @ap.cmd(
        ap.arg('roots', nargs='+'),
    )
    def gen(self) -> None:
        gp = GenPy(
            self.args.roots,
        )
        gp.run()


def _main() -> None:
    logs.configure_standard_logging(logging.INFO)
    cli = Cli()
    cli()


if __name__ == '__main__':
    _main()



# type: ignore
# ruff: noqa
# flake8: noqa
# @omlish-generated
# Generated from Json.g4 by ANTLR 4.13.2
from ....text.antlr._runtime._all import *
if "." in __name__:
    from .JsonParser import JsonParser
else:
    from JsonParser import JsonParser

# This class defines a complete generic visitor for a parse tree produced by JsonParser.

class JsonVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by JsonParser#json.
    def visitJson(self, ctx:JsonParser.JsonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by JsonParser#obj.
    def visitObj(self, ctx:JsonParser.ObjContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by JsonParser#pair.
    def visitPair(self, ctx:JsonParser.PairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by JsonParser#arr.
    def visitArr(self, ctx:JsonParser.ArrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by JsonParser#value.
    def visitValue(self, ctx:JsonParser.ValueContext):
        return self.visitChildren(ctx)



del JsonParser


import dataclasses as dc
import typing as ta

from omlish import check
from omlish import lang


FormatPart: ta.TypeAlias = ta.Union[str, 'Formats.Optional', 'Formats.Variadic']
FormatParts: ta.TypeAlias = ta.Sequence[FormatPart]


##


class Formats(lang.Namespace):
    @dc.dataclass(frozen=True)
    class Name:
        name: str

    @dc.dataclass(frozen=True)
    class Optional:
        body: FormatParts

    @dc.dataclass(frozen=True)
    class Variadic:
        body: FormatParts

    #

    _PARTS_BY_DELIMITERS: ta.Mapping[tuple[str, str], type] = {
        ('[', ']'): Optional,
        ('{', '}'): Variadic,
    }

    _DELIMITERS_BY_PARTS: ta.Mapping[type, tuple[str, str]] = {v: k for k, v in _PARTS_BY_DELIMITERS.items()}

    #

    @staticmethod
    def split_parts(s: str) -> FormatParts:
        stk: list[tuple[str, list]] = [('', [])]

        p = 0
        while p < len(s):
            n = lang.find_any(s, '{}[]<', p)

            if n < 0:
                check.state(not stk[-1][0])
                stk[-1][1].append(s[p:])
                break

            if n != p:
                stk[-1][1].append(s[p:n])

            d = s[n]
            if d == '<':
                e = s.index('>', n)
                stk[-1][1].append(Formats.Name(s[n + 1:e]))
                p = e + 1

            elif d in '{[':
                stk.append((d, []))
                p = n + 1

            elif d in '}]':
                x, l = stk.pop()
                pc = Formats._PARTS_BY_DELIMITERS[(x, d)]
                stk[-1][1].append(pc(l))
                p = n + 1

            else:
                raise RuntimeError

        _, ret = check.single(stk)
        return ret

    #

    @staticmethod
    def render_parts(p: FormatPart | FormatParts) -> ta.Iterator[str]:
        if isinstance(p, str):
            yield p

        elif isinstance(p, Formats.Name):
            yield '<'
            yield p.name
            yield '>'

        elif isinstance(p, (Formats.Optional, Formats.Variadic)):
            l, r = Formats._DELIMITERS_BY_PARTS[type(p)]
            yield l
            yield from Formats.render_parts(p.body)
            yield r

        else:
            for c in p:
                yield from Formats.render_parts(c)



import abc
import contextlib
import io
import typing as ta

from omlish import check
from omlish import lang
from omlish.concurrent import threadlets as tls
from omlish.sync import ConditionDeque


if ta.TYPE_CHECKING:
    import threading

    from omlish.io import pyio  # noqa

else:
    threading = lang.proxy_import('threading')

    pyio = lang.proxy_import('omlish.io.pyio')


T = ta.TypeVar('T')

BytesLike: ta.TypeAlias = ta.Any

BufferedReader: ta.TypeAlias = io.BufferedReader
# BufferedReader: ta.TypeAlias = pyio.BufferedReader


##


class ProxyReadFile:
    def __init__(self, read: ta.Callable[[int], bytes | BaseException]) -> None:
        super().__init__()

        self._read = read
        self._eof = False
        self._closed = False

    #

    def read(self, n: int, /) -> bytes:
        if self._closed:
            raise RuntimeError('Closed')
        if self._eof:
            return b''
        d = self._read(n)
        if isinstance(d, BaseException):
            raise d
        if not d:
            self._eof = True
        return d

    def readall(self, *args, **kwargs):
        raise TypeError  # FIXME

    def readinto(self, b: BytesLike) -> int | None:
        d = self.read(len(b))
        if d:
            b[:len(d)] = d
        return len(d)

    #

    def readable(self) -> bool:
        return not (self._eof or self._closed)

    @property
    def closed(self) -> bool:
        return self._closed

    def close(self) -> None:
        self._closed = True

    #

    def flush(self) -> None:
        pass

    def seekable(self) -> bool:
        return False

    def seek(self, n: int, /) -> ta.Any:
        raise TypeError


##


class NeedMore(lang.Marker):
    pass


class Exited(lang.Marker):
    pass


class Shutdown(BaseException):  # noqa
    pass


IoTrampolineTarget: ta.TypeAlias = ta.Callable[[io.BufferedReader], ta.ContextManager[ta.Callable[[], bytes]]]


class IoTrampoline(lang.Abstract):
    def __init__(
            self,
            target: IoTrampolineTarget,
            *,
            buffer_size: int | None = None,
    ) -> None:
        super().__init__()

        self._target = target
        self._buffer_size = buffer_size

    def _make_buffered_reader(self, raw: ta.Any) -> io.BufferedReader:
        return io.BufferedReader(
            raw,
            **(dict(buffer_size=self._buffer_size) if self._buffer_size is not None else {}),
        )

    @abc.abstractmethod
    def close(self, timeout: float | None = None) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def __enter__(self) -> ta.Self:
        raise NotImplementedError

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    @abc.abstractmethod
    def feed(self, *data: bytes) -> ta.Iterator[bytes]:
        raise NotImplementedError


#


class ThreadIoTrampoline(IoTrampoline):
    def __init__(self, target: IoTrampolineTarget, **kwargs: ta.Any) -> None:
        super().__init__(target, **kwargs)

        self._in: ConditionDeque[bytes | BaseException] = ConditionDeque()
        self._out: ConditionDeque[
            bytes |  # noqa
            type[NeedMore] |
            type[Exited] |
            BaseException
        ] = ConditionDeque()

        self._proxy = ProxyReadFile(self._read)

        self._thread = threading.Thread(target=self._thread_proc, daemon=True)

    #

    def close(self, timeout: float | None = None) -> None:
        if not self._thread.is_alive():
            return
        self._out.push(Shutdown())
        self._thread.join(timeout)
        if self._thread.is_alive():
            if timeout is not None:
                raise TimeoutError
            else:
                raise RuntimeError('Failed to join thread')

    #

    def __enter__(self) -> ta.Self:
        self._thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    #

    def _read(self, n: int, /) -> bytes | BaseException:
        return self._in.pop(if_empty=lambda: self._out.push(NeedMore))

    def _thread_proc(self) -> None:
        try:
            with contextlib.closing(self._make_buffered_reader(self._proxy)) as bf:  # noqa
                with self._target(bf) as read:
                    while out := read():
                        self._out.push(out)
                    self._out.push(out)
        except BaseException as e:
            self._out.push(e)
            raise
        finally:
            self._out.push(Exited)

    def feed(self, *data: bytes) -> ta.Iterator[bytes]:
        self._in.push(*data)
        while True:
            e = self._out.pop()
            if e is NeedMore:
                break
            elif isinstance(e, BaseException):
                raise e
            elif e is Exited:
                raise RuntimeError('IO thread exited')
            elif isinstance(e, bytes):
                yield e
                if not e:
                    return
            else:
                raise TypeError(e)


#


class ThreadletIoTrampoline(IoTrampoline):
    def __init__(
            self,
            target: IoTrampolineTarget,
            threadlets: tls.Threadlets = tls.GREENLET_THREADLETS,
            ** kwargs: ta.Any,
    ) -> None:
        super().__init__(target, **kwargs)

        self._proxy = ProxyReadFile(self._read)
        self._tl: tls.Threadlet = threadlets.spawn(self._g_proc)

    #

    def close(self, timeout: float | None = None) -> None:
        if self._tl.dead:
            return
        out = self._tl.switch(Shutdown())
        if out is not Exited or not self._tl.dead:
            raise RuntimeError

    #

    def __enter__(self) -> ta.Self:
        out = self._tl.switch()
        if out is not NeedMore:
            raise RuntimeError
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    #

    def _read(self, n: int, /) -> bytes:
        out = check.not_none(self._tl.parent).switch(NeedMore)
        return out

    def _g_proc(self) -> ta.Any:
        try:
            with contextlib.closing(self._make_buffered_reader(self._proxy)) as bf:  # noqa
                with self._target(bf) as read:
                    while out := read():
                        e = check.not_none(self._tl.parent).switch(out)
                        if e is not NeedMore:
                            raise TypeError(e)  # noqa
                    e = check.not_none(self._tl.parent).switch(out)
                    if not isinstance(e, Shutdown):
                        raise TypeError(e)  # noqa
            return Exited
        except BaseException as e:
            check.not_none(self._tl.parent).throw(e)
            raise

    def feed(self, *data: bytes) -> ta.Iterator[bytes]:
        i: bytes | type[NeedMore]
        for i in data:
            while True:
                e = self._tl.switch(i)
                i = NeedMore
                if e is NeedMore:
                    break
                elif isinstance(e, bytes):
                    yield e
                    if not e:
                        return
                else:
                    raise TypeError(e)



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#

from io import StringIO
from .Token import Token

# need forward declarations
IntervalSet = None

class IntervalSet(object):
    __slots__ = ('intervals', 'readonly')

    def __init__(self):
        self.intervals = None
        self.readonly = False

    def __iter__(self):
        if self.intervals is not None:
            for i in self.intervals:
                for c in i:
                    yield c

    def __getitem__(self, item):
        i = 0
        for k in self:
            if i==item:
                return k
            else:
                i += 1
        return Token.INVALID_TYPE

    def addOne(self, v:int):
        self.addRange(range(v, v+1))

    def addRange(self, v:range):
        if self.intervals is None:
            self.intervals = list()
            self.intervals.append(v)
        else:
            # find insert pos
            k = 0
            for i in self.intervals:
                # distinct range -> insert
                if v.stop<i.start:
                    self.intervals.insert(k, v)
                    return
                # contiguous range -> adjust
                elif v.stop==i.start:
                    self.intervals[k] = range(v.start, i.stop)
                    return
                # overlapping range -> adjust and reduce
                elif v.start<=i.stop:
                    self.intervals[k] = range(min(i.start,v.start), max(i.stop,v.stop))
                    self.reduce(k)
                    return
                k += 1
            # greater than any existing
            self.intervals.append(v)

    def addSet(self, other:IntervalSet):
        if other.intervals is not None:
            for i in other.intervals:
                self.addRange(i)
        return self

    def reduce(self, k:int):
        # only need to reduce if k is not the last
        if k<len(self.intervals)-1:
            l = self.intervals[k]
            r = self.intervals[k+1]
            # if r contained in l
            if l.stop >= r.stop:
                self.intervals.pop(k+1)
                self.reduce(k)
            elif l.stop >= r.start:
                self.intervals[k] = range(l.start, r.stop)
                self.intervals.pop(k+1)

    def complement(self, start, stop):
        result = IntervalSet()
        result.addRange(range(start,stop+1))
        for i in self.intervals:
            result.removeRange(i)
        return result

    def __contains__(self, item):
        if self.intervals is None:
            return False
        else:
            return any(item in i for i in self.intervals)

    def __len__(self):
        return sum(len(i) for i in self.intervals)

    def removeRange(self, v):
        if v.start==v.stop-1:
            self.removeOne(v.start)
        elif self.intervals is not None:
            k = 0
            for i in self.intervals:
                # intervals are ordered
                if v.stop<=i.start:
                    return
                # check for including range, split it
                elif v.start>i.start and v.stop<i.stop:
                    self.intervals[k] = range(i.start, v.start)
                    x = range(v.stop, i.stop)
                    self.intervals.insert(k, x)
                    return
                # check for included range, remove it
                elif v.start<=i.start and v.stop>=i.stop:
                    self.intervals.pop(k)
                    k -= 1  # need another pass
                # check for lower boundary
                elif v.start<i.stop:
                    self.intervals[k] = range(i.start, v.start)
                # check for upper boundary
                elif v.stop<i.stop:
                    self.intervals[k] = range(v.stop, i.stop)
                k += 1

    def removeOne(self, v):
        if self.intervals is not None:
            k = 0
            for i in self.intervals:
                # intervals is ordered
                if v<i.start:
                    return
                # check for single value range
                elif v==i.start and v==i.stop-1:
                    self.intervals.pop(k)
                    return
                # check for lower boundary
                elif v==i.start:
                    self.intervals[k] = range(i.start+1, i.stop)
                    return
                # check for upper boundary
                elif v==i.stop-1:
                    self.intervals[k] = range(i.start, i.stop-1)
                    return
                # split existing range
                elif v<i.stop-1:
                    x = range(i.start, v)
                    self.intervals[k] = range(v + 1, i.stop)
                    self.intervals.insert(k, x)
                    return
                k += 1


    def toString(self, literalNames:list, symbolicNames:list):
        if self.intervals is None:
            return "{}"
        with StringIO() as buf:
            if len(self)>1:
                buf.write("{")
            first = True
            for i in self.intervals:
                for j in i:
                    if not first:
                        buf.write(", ")
                    buf.write(self.elementName(literalNames, symbolicNames, j))
                    first = False
            if len(self)>1:
                buf.write("}")
            return buf.getvalue()

    def elementName(self, literalNames:list, symbolicNames:list, a:int):
        if a==Token.EOF:
            return "<EOF>"
        elif a==Token.EPSILON:
            return "<EPSILON>"
        else:
            if a<len(literalNames) and literalNames[a] != "<INVALID>":
                return literalNames[a]
            if a<len(symbolicNames):
                return symbolicNames[a]
            return "<UNKNOWN>"



import enum
import typing as ta

from omlish import check
from omlish import dataclasses as dc
from omlish import lang
from omlish.funcs import pairs as fps


MessageParamsUnpacker: ta.TypeAlias = fps.FnPair[
    ta.Mapping[str, str],  # params
    ta.Mapping[str, ta.Any],  # kwargs
]


##


class MessageFormat(dc.Frozen, final=True):
    name: str

    class Param(dc.Case):
        @classmethod
        def of(cls, obj: ta.Any) -> 'MessageFormat.Param':
            if isinstance(obj, MessageFormat.Param):
                return obj

            elif isinstance(obj, str):
                s = check.non_empty_str(obj)

                optional = False
                if s.startswith('?'):
                    optional = True
                    s = s[1:]

                arity = MessageFormat.KwargParam.Arity.SINGLE
                if s.startswith('*'):
                    arity = MessageFormat.KwargParam.Arity.VARIADIC
                    s = s[1:]

                elif s.startswith(','):
                    arity = MessageFormat.KwargParam.Arity.COMMA_LIST
                    s = s[1:]

                return MessageFormat.KwargParam(
                    s,
                    optional=optional,
                    arity=arity,
                )

            else:
                raise TypeError(obj)

    class KwargParam(Param):
        name: str = dc.xfield(validate=lang.is_ident)

        optional: bool = False

        class Arity(enum.Enum):
            SINGLE = enum.auto()  # <foo>
            VARIADIC = enum.auto()  # <foo>{ <foo>}
            COMMA_LIST = enum.auto()  # <foo>{,<foo>}

        arity: Arity = Arity.SINGLE

    class LiteralParam(Param):
        text: str

    params: ta.Sequence[Param]

    _: dc.KW_ONLY

    unpack_params: MessageParamsUnpacker | None = None

    @dc.init
    def _validate_params(self) -> None:
        kws = [p for p in self.params if isinstance(p, MessageFormat.KwargParam)]
        check.unique(p.name for p in kws)
        check.state(all(p.arity is not MessageFormat.KwargParam.Arity.VARIADIC for p in kws[:-1]))

    @classmethod
    def of(
            cls,
            name: str,
            *params: ta.Any,
            **kwargs: ta.Any,
    ) -> 'MessageFormat':
        return cls(
            name,
            [MessageFormat.Param.of(p) for p in params],
            **kwargs,
        )



"""
TODO:
 - mtime cmp
 - parallelism
"""
import os.path
import re
import shutil
import subprocess
import typing as ta

from omlish import check
from omlish import lang
from omlish.logs import all as logs
from omlish.os.paths import is_path_in_dir

from .consts import ANTLR_JAR_URL
from .consts import ANTLR_RUNTIME_VENDOR


if ta.TYPE_CHECKING:
    from omdev.cache import data as dcache
else:
    dcache = lang.proxy_import('omdev.cache.data')


log = logs.get_module_logger(globals())


##


ANTLR_JAR_CACHE = dcache.UrlSpec(ANTLR_JAR_URL)


@lang.cached_function
def get_jar_path() -> str:
    return dcache.default().get(ANTLR_JAR_CACHE)


##


def _find_dirs(*base_paths: str, filter: ta.Callable[[str], bool] = lambda _: True) -> ta.Sequence[str]:  # noqa
    return sorted(
        os.path.join(dp, dn)
        for base_path in base_paths
        for dp, dns, fns in os.walk(base_path)
        for dn in dns
        if filter(dn)
    )


def _find_files(*base_paths: str, filter: ta.Callable[[str], bool] = lambda _: True) -> ta.Sequence[str]:  # noqa
    return sorted(
        os.path.join(dp, fn)
        for base_path in base_paths
        for dp, dns, fns in os.walk(base_path)
        for fn in fns
        if filter(fn)
    )


class GenPy:
    def __init__(
            self,
            root_dirs: str,  # noqa
            *,
            out_subdir: str = '_antlr',
            runtime_import: str = ANTLR_RUNTIME_VENDOR,
            jar_path: str | None = None,
            # parallelism: int | None = None,
    ) -> None:
        super().__init__()

        check.non_empty_str(out_subdir)
        check.arg(not os.path.isabs(out_subdir) and '.' not in out_subdir and '/' not in out_subdir)

        self._root_dirs = frozenset(check.non_empty_str(rd) for rd in check.not_isinstance(root_dirs, str))
        self._out_subdir = out_subdir
        self._runtime_import = runtime_import
        self._given_jar_path = jar_path

    #

    def _rmtree(self, tgt: str) -> None:  # noqa
        if not any(is_path_in_dir(rd, tgt) for rd in self._root_dirs):
            raise RuntimeError(f'Refusing to delete {tgt!r} outside of {self._root_dirs!r}')
        shutil.rmtree(tgt)

    #

    @lang.cached_function
    def jar_path(self) -> str:
        if (gjp := self._given_jar_path) is not None:
            return gjp
        return get_jar_path()

    #

    def process_g4(self, g4_file: str) -> None:
        ap = os.path.abspath(g4_file)
        check.state(os.path.isfile(ap))

        od = os.path.join(os.path.dirname(ap), self._out_subdir)
        os.makedirs(od, exist_ok=True)

        log.info('Compiling grammar %s', g4_file)

        try:
            subprocess.check_call([
                'java',
                '-jar', self.jar_path(),
                '-Dlanguage=Python3',
                '-visitor',
                '-o', self._out_subdir,
                os.path.basename(g4_file),
            ], cwd=os.path.dirname(ap))

        except Exception:  # noqa
            log.exception('Exception in grammar %s', g4_file)
            raise

    def process_py(self, py_file: str) -> None:
        ap = os.path.abspath(py_file)
        with open(ap) as f:
            in_lines = list(f)

        pfp = py_file.split(os.sep)
        arp = ANTLR_RUNTIME_VENDOR.split('.')
        if (cpl := lang.common_prefix_len(pfp, arp)) > 0:
            pkg_depth = len(os.path.normpath(py_file).split(os.path.sep))
            antlr_imp = '.'.join([*([''] * (pkg_depth - cpl)), *arp[cpl:]])
        else:
            antlr_imp = ANTLR_RUNTIME_VENDOR

        out_lines = [
            '# type: ignore\n',
            '# ruff: noqa\n',
            '# flake8: noqa\n',
            '# @omlish-generated\n',
        ]

        for l in in_lines:
            l = re.sub(r'^(from antlr4)(.*)', rf'from {antlr_imp}\2', l)
            out_lines.append(l)

        with open(ap, 'w') as f:
            f.write(''.join(out_lines))

    def process_dir(self, dir: str) -> None:  # noqa
        log.info('Processing directory %s', dir)

        ad = os.path.join(dir, self._out_subdir)
        if os.path.exists(ad):
            self._rmtree(ad)

        for f in os.listdir(dir):
            fp = os.path.join(dir, f)
            if not os.path.isfile(fp) or not f.endswith('.g4'):
                continue

            self.process_g4(fp)

        if not os.path.exists(ad):
            return

        ip = os.path.join(ad, '__init__.py')
        check.state(not os.path.exists(ip))

        for f in list(os.listdir(ad)):
            fp = os.path.join(ad, f)
            if not os.path.isfile(fp):
                continue

            if f.split('.')[-1] in ('interp', 'tokens'):
                os.unlink(fp)

            elif f != '__init__.py' and f.endswith('.py'):
                self.process_py(fp)

        with open(ip, 'w'):
            pass

    def run(self) -> None:
        dns = _find_dirs(*self._root_dirs, filter=lambda dn: os.path.basename(dn) == '_antlr')
        for dn in dns:
            self._rmtree(dn)

        fns = _find_files(*self._root_dirs, filter=lambda fn: fn.endswith('.g4'))
        fds = {os.path.dirname(fn) for fn in fns}
        for dn in sorted(fds):
            self.process_dir(dn)



"""
TODO:
 - PONG - optional *before* required ...
 - CAPABILITIES - ':' last param

==

https://datatracker.ietf.org/doc/html/rfc2812
https://modern.ircdocs.horse/
"""
import typing as ta

from omlish import check
from omlish import dataclasses as dc

from ..numerics import numerics as nr
from .base import Message
from .base import MessageFormat
from .base import list_pair_params_unpacker


##


_REGISTERED_IRC_MESSAGES_BY_NAME: dict[str, list[type['Message']]] = {}


def _register_irc_message(cls):
    check.issubclass(cls, Message)
    _REGISTERED_IRC_MESSAGES_BY_NAME.setdefault(cls.FORMAT.name, []).append(cls)
    return cls


##
# Connection Messages


@_register_irc_message
class CapMessage(Message):
    FORMAT = MessageFormat.of(
        'CAP',
        'subcommand',
        'capabilities',
    )

    subcommand: str
    capabilities: ta.Sequence[str]


@_register_irc_message
class AuthenticateMessage(Message):
    FORMAT = MessageFormat.of('AUTHENTICATE')


@_register_irc_message
class PassMessage(Message):
    FORMAT = MessageFormat.of(
        'PASS',
        'password',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_ALREADYREGISTERED,
        nr.ERR_PASSWDMISMATCH,
    )

    password: str


@_register_irc_message
class NickMessage(Message):
    FORMAT = MessageFormat.of(
        'NICK',
        'nickname',
    )

    REPLIES = (
        nr.ERR_NONICKNAMEGIVEN,
        nr.ERR_ERRONEUSNICKNAME,
        nr.ERR_NICKNAMEINUSE,
        nr.ERR_NICKCOLLISION,
    )

    nickname: str


class UserMessage(Message):
    FORMAT = MessageFormat.of(
        'USER',
        'username',
        'param1',  # 0
        'param2',  # *
        'realname',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_ALREADYREGISTERED,
    )

    username: str
    realname: str

    param1: str = '8'
    param2: str = '*'


@_register_irc_message
class PingMessage(Message):
    FORMAT = MessageFormat.of(
        'PING',
        'token',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOORIGIN,
        nr.ERR_NOSUCHSERVER,
    )

    token: str


@_register_irc_message
class PongMessage(Message):
    FORMAT = MessageFormat.of(
        'PONG',
        '?server',
        'token',
    )

    token: str
    server: str | None = None


@_register_irc_message
class OperMessage(Message):
    FORMAT = MessageFormat.of(
        'OPER',
        'name',
        'password',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_PASSWDMISMATCH,
        nr.ERR_NOOPERHOST,
        nr.RPL_YOUREOPER,
    )

    name: str
    password: str


@_register_irc_message
class QuitMessage(Message):
    FORMAT = MessageFormat.of(
        'QUIT',
        '?reason',
    )

    reason: str | None = None


@_register_irc_message
class ErrorMessage(Message):
    FORMAT = MessageFormat.of(
        'ERROR',
        '?reason',
    )

    reason: str | None = None


##
# Channel Operations


@_register_irc_message
class JoinMessage(Message):
    FORMAT = MessageFormat.of(
        'JOIN',
        ','
        'channels', '?,keys',
        unpack_params=list_pair_params_unpacker(
            'channels',
            'channels',
            'keys',
        ),
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOSUCHCHANNEL,
        nr.ERR_TOOMANYCHANNELS,
        nr.ERR_BADCHANNELKEY,
        nr.ERR_BANNEDFROMCHAN,
        nr.ERR_CHANNELISFULL,
        nr.ERR_INVITEONLYCHAN,
        nr.ERR_BADCHANMASK,
        nr.RPL_TOPIC,
        nr.RPL_TOPICWHOTIME,
        nr.RPL_NAMREPLY,
        nr.RPL_ENDOFNAMES,
    )

    channels: ta.Sequence[str | tuple[str, str]]

    @property
    def has_keys(self) -> bool:
        return check.single({isinstance(e, tuple) for e in self.channels})

    @dc.init
    def _validate_channels(self) -> None:
        check.not_isinstance(self.channels, str)
        if self.has_keys:
            for c, _ in self.channels:  # type: ignore[misc]
                check.non_empty_str(c)
        else:
            for c in self.channels:
                check.non_empty_str(c)  # type: ignore


@_register_irc_message
class LeaveAllJoinMessage(Message):
    FORMAT = MessageFormat.of(
        'JOIN',
        MessageFormat.LiteralParam('0'),
    )

    REPLIES = JoinMessage.REPLIES


@_register_irc_message
class PartMessage(Message):
    FORMAT = MessageFormat.of(
        'PART',
        ',channels',
        '?reason',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOSUCHCHANNEL,
        nr.ERR_NOTONCHANNEL,
    )

    channels: ta.Sequence[str] = dc.xfield(validate=check.of_not_isinstance(str))
    reason: str | None = None


@_register_irc_message
class TopicMessage(Message):
    FORMAT = MessageFormat.of(
        'TOPIC',
        'channel',
        '?topic',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOSUCHCHANNEL,
        nr.ERR_NOTONCHANNEL,
        nr.ERR_CHANOPRIVSNEEDED,
        nr.RPL_NOTOPIC,
        nr.RPL_TOPIC,
        nr.RPL_TOPICWHOTIME,
    )

    channel: str
    topic: str | None = None


@_register_irc_message
class NamesMessage(Message):
    FORMAT = MessageFormat.of(
        'NAMES',
        ',channels',
    )

    REPLIES = (
        nr.RPL_NAMREPLY,
        nr.RPL_ENDOFNAMES,
    )

    channels: ta.Sequence[str] = dc.xfield(validate=check.of_not_isinstance(str))


@_register_irc_message
class ListMessage(Message):
    FORMAT = MessageFormat.of(
        'LIST',
        '?,channels',
        '?,elistconds',
    )

    REPLIES = (
        nr.RPL_LISTSTART,
        nr.RPL_LIST,
        nr.RPL_LISTEND,
    )

    channels: ta.Sequence[str]
    elistconds: ta.Sequence[str]


@_register_irc_message
class InviteMessage(Message):
    FORMAT = MessageFormat.of(
        'INVITE',
        'nickname',
        'channel',
    )

    REPLIES = (
        nr.RPL_INVITING,
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOSUCHCHANNEL,
        nr.ERR_NOTONCHANNEL,
        nr.ERR_CHANOPRIVSNEEDED,
        nr.ERR_USERONCHANNEL,
    )

    nickname: str
    channel: str


@_register_irc_message
class KickMessage(Message):
    FORMAT = MessageFormat.of(
        'KICK',
        'channel',
        ',users',
        '?comment',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOSUCHCHANNEL,
        nr.ERR_CHANOPRIVSNEEDED,
        nr.ERR_USERNOTINCHANNEL,
        nr.ERR_NOTONCHANNEL,
        nr.ERR_BADCHANMASK,
    )

    channel: str
    users: ta.Sequence[str]
    comment: str | None = None


##
# Server Queries and Commands


@_register_irc_message
class MotdMessage(Message):
    FORMAT = MessageFormat.of(
        'MOTD',
        '?target',
    )

    REPLIES = (
        nr.ERR_NOSUCHSERVER,
        nr.ERR_NOMOTD,
        nr.RPL_MOTDSTART,
        nr.RPL_MOTD,
        nr.RPL_ENDOFMOTD,
    )

    target: str | None = None


@_register_irc_message
class VersionMessage(Message):
    FORMAT = MessageFormat.of(
        'VERSION',
        '?target',
    )

    REPLIES = (
        nr.ERR_NOSUCHSERVER,
        nr.RPL_ISUPPORT,
        nr.RPL_VERSION,
    )

    target: str | None = None


@_register_irc_message
class AdminMessage(Message):
    FORMAT = MessageFormat.of(
        'ADMIN',
        '?target',
    )

    REPLIES = (
        nr.ERR_NOSUCHSERVER,
        nr.RPL_ADMINME,
        nr.RPL_ADMINLOC1,
        nr.RPL_ADMINLOC2,
        nr.RPL_ADMINEMAIL,
    )

    target: str | None = None


@_register_irc_message
class ConnectMessage(Message):
    FORMAT = MessageFormat.of(
        'CONNECT',
        'target_server',
        '?port',
        '?remote_server',
    )

    REPLIES = (
        nr.ERR_NOSUCHSERVER,
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOPRIVILEGES,
        nr.ERR_NOPRIVS,
    )

    target_server: str
    port: str | None = None
    remote_server: str | None = None


@_register_irc_message
class LusersMessage(Message):
    FORMAT = MessageFormat.of('LUSERS')

    REPLIES = (
        nr.RPL_LUSERCLIENT,
        nr.RPL_LUSEROP,
        nr.RPL_LUSERUNKNOWN,
        nr.RPL_LUSERCHANNELS,
        nr.RPL_LUSERME,
        nr.RPL_LOCALUSERS,
        nr.RPL_GLOBALUSERS,
    )


@_register_irc_message
class TimeMessage(Message):
    FORMAT = MessageFormat.of(
        'TIME',
        '?server',
    )

    REPLIES = (
        nr.ERR_NOSUCHSERVER,
        nr.RPL_TIME,
    )

    server: str | None = None


@_register_irc_message
class StatsMessage(Message):
    FORMAT = MessageFormat.of(
        'STATS',
        'query',
        '?server',
    )

    REPLIES = (
        nr.ERR_NOSUCHSERVER,
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOPRIVILEGES,
        nr.ERR_NOPRIVS,
        nr.RPL_STATSCLINE,
        nr.RPL_STATSHLINE,
        nr.RPL_STATSILINE,
        nr.RPL_STATSKLINE,
        nr.RPL_STATSLLINE,
        nr.RPL_STATSOLINE,
        nr.RPL_STATSLINKINFO,
        nr.RPL_STATSUPTIME,
        nr.RPL_STATSCOMMANDS,
        nr.RPL_ENDOFSTATS,
    )

    query: str
    server: str | None = None


@_register_irc_message
class HelpMessage(Message):
    FORMAT = MessageFormat.of(
        'HELP',
        '?subject',
    )

    REPLIES = (
        nr.ERR_HELPNOTFOUND,
        nr.RPL_HELPSTART,
        nr.RPL_HELPTXT,
        nr.RPL_ENDOFHELP,
    )

    subject: str | None = None


@_register_irc_message
class InfoMessage(Message):
    FORMAT = MessageFormat.of('INFO')

    REPLIES = (
        nr.RPL_INFO,
        nr.RPL_ENDOFINFO,
    )


@_register_irc_message
class ModeMessage(Message):
    FORMAT = MessageFormat.of(
        'MODE',
        'target',
        '?modestring',
        '*mode_arguments',
    )

    REPLIES = (
        nr.ERR_NOSUCHCHANNEL,
        nr.RPL_CHANNELMODEIS,
        nr.RPL_CREATIONTIME,
        nr.RPL_CHANNELMODEIS,
        nr.ERR_CHANOPRIVSNEEDED,
        nr.RPL_BANLIST,
        nr.RPL_ENDOFBANLIST,
        nr.RPL_EXCEPTLIST,
        nr.RPL_ENDOFEXCEPTLIST,
        nr.RPL_INVITELIST,
        nr.RPL_ENDOFINVITELIST,
    )

    target: str
    modestring: str | None = None
    mode_arguments: ta.Sequence[str] | None = None


##
# Sending Messages


@_register_irc_message
class PrivmsgMessage(Message):
    FORMAT = MessageFormat.of(
        'PRIVMSG',
        ',targets',
        'text',
    )

    REPLIES = (
        nr.ERR_NOSUCHNICK,
        nr.ERR_NOSUCHSERVER,
        nr.ERR_CANNOTSENDTOCHAN,
        nr.ERR_TOOMANYTARGETS,
        nr.ERR_NORECIPIENT,
        nr.ERR_NOTEXTTOSEND,
        nr.ERR_NOTOPLEVEL,
        nr.ERR_WILDTOPLEVEL,
        nr.RPL_AWAY,
    )

    targets: ta.Sequence[str]
    text: str


@_register_irc_message
class NoticeMessage(Message):
    FORMAT = MessageFormat.of(
        'NOTICE',
        ',targets',
        'text',
    )

    targets: ta.Sequence[str]
    text: str


##
# User-Based Queries

@_register_irc_message
class WhoMessage(Message):
    FORMAT = MessageFormat.of(
        'WHO',
        'mask',
    )

    REPLIES = (
        nr.RPL_WHOREPLY,
        nr.RPL_ENDOFWHO,
    )

    mask: str


@_register_irc_message
class WhoisMessage(Message):
    FORMAT = MessageFormat.of(
        'WHOIS',
        '?target',
        'nick',
    )

    REPLIES = (
        nr.ERR_NOSUCHNICK,
        nr.ERR_NOSUCHSERVER,
        nr.ERR_NONICKNAMEGIVEN,
        nr.RPL_WHOISCERTFP,
        nr.RPL_WHOISREGNICK,
        nr.RPL_WHOISUSER,
        nr.RPL_WHOISSERVER,
        nr.RPL_WHOISOPERATOR,
        nr.RPL_WHOISIDLE,
        nr.RPL_WHOISCHANNELS,
        nr.RPL_WHOISSPECIAL,
        nr.RPL_WHOISACCOUNT,
        nr.RPL_WHOISACTUALLY,
        nr.RPL_WHOISHOST,
        nr.RPL_WHOISMODES,
        nr.RPL_WHOISSECURE,
        nr.RPL_AWAY,
    )

    nick: str
    target: str | None = None


@_register_irc_message
class WhowasMessage(Message):
    FORMAT = MessageFormat.of(
        'WHOWAS',
        'nick',
        '?count',
    )

    REPLIES = (
        nr.ERR_WASNOSUCHNICK,
        nr.RPL_ENDOFWHOWAS,
        nr.RPL_WHOWASUSER,
        nr.RPL_WHOISACTUALLY,
        nr.RPL_WHOISSERVER,
        nr.RPL_WHOWASUSER,
        nr.ERR_NONICKNAMEGIVEN,
        nr.ERR_NEEDMOREPARAMS,
    )

    nick: str
    count: str | None = None


##
# Operator Messages


@_register_irc_message
class KillMessage(Message):
    FORMAT = MessageFormat.of(
        'KILL',
        'nickname',
        'comment',
    )

    REPLIES = (
        nr.ERR_NOSUCHSERVER,
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOPRIVILEGES,
        nr.ERR_NOPRIVS,
    )

    nickname: str
    comment: str


@_register_irc_message
class RehashMessage(Message):
    FORMAT = MessageFormat.of('REHASH')

    REPLIES = (
        nr.RPL_REHASHING,
        nr.ERR_NOPRIVILEGES,
    )


@_register_irc_message
class RestartMessage(Message):
    FORMAT = MessageFormat.of('RESTART')

    REPLIES = (
        nr.ERR_NOPRIVILEGES,
    )


@_register_irc_message
class SquitMessage(Message):
    FORMAT = MessageFormat.of(
        'SQUIT',
        'server',
        'comment',
    )

    REPLIES = (
        nr.ERR_NOSUCHSERVER,
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOPRIVILEGES,
        nr.ERR_NOPRIVS,
    )

    server: str
    comment: str


##
# Optional Messages


@_register_irc_message
class AwayMessage(Message):
    FORMAT = MessageFormat.of(
        'AWAY',
        '?text',
    )

    REPLIES = (
        nr.RPL_AWAY,
        nr.RPL_UNAWAY,
        nr.RPL_NOWAWAY,
    )

    text: str | None = None


@_register_irc_message
class LinksMessage(Message):
    FORMAT = MessageFormat.of('LINKS')

    REPLIES = (
        nr.RPL_LINKS,
        nr.RPL_ENDOFLINKS,
    )


@_register_irc_message
class UserhostMessage(Message):
    FORMAT = MessageFormat.of(
        'USERHOST',
        '*nicknames',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.RPL_USERHOST,
    )

    nicknames: ta.Sequence[str] | None = dc.xfield(default=None, validate=check.of_not_isinstance(str))


@_register_irc_message
class WallopsMessage(Message):
    FORMAT = MessageFormat.of(
        'WALLOPS',
        'text',
    )

    REPLIES = (
        nr.ERR_NEEDMOREPARAMS,
        nr.ERR_NOPRIVILEGES,
        nr.ERR_NOPRIVS,
    )

    text: str



ANTLR_VERSION = '4.13.2'
ANTLR_JAR_NAME = f'antlr-{ANTLR_VERSION}-complete.jar'
ANTLR_JAR_URL = f'https://www.antlr.org/download/{ANTLR_JAR_NAME}'
ANTLR_RUNTIME_PACKAGE = 'antlr4-python3-runtime'
ANTLR_GITHUB_REPO = 'antlr/antlr4'

ANTLR_RUNTIME_VENDOR = 'omextra.text.antlr._runtime._all'
